
#include "PrintTools.h"

void check_matrix_dimensions(int length,int rows)
{
	if((length%rows)!=0)
	{
		print_error("Argument is not a matrix, check dimensions",__func__,"Aborting...");
		exit(1);
	}
}

void print_info(const char* info)
{
	printf(ANSI_COLOR_GREEN"[i] <info>: "ANSI_COLOR_RESET);
	printf(info); printf("\n");
}
void print_warning(const char* why, const char* where,const char* what)
{
	printf(ANSI_COLOR_YELLOW"[!] <warning>: "ANSI_COLOR_RESET);
	printf(why); printf(". In <"); printf(where); printf(">. "); printf(what);  printf(".\n");
}
void print_error(const char* why, const char* where,const char* what)
{
	printf(ANSI_COLOR_RED"[x] <error>: "ANSI_COLOR_RESET);
	printf(why); printf(". In <"); printf(where); printf(">. "); printf(what);  printf(".\n");
}

void print_matrix(const char *mname,const double* matrix, int length, int rows,int layout)
{
	printf("MATRIX: %s > ", mname );
	check_matrix_dimensions(length,rows);

	if(layout==COLUMN_MAJOR)   printf("column-major layout (FORTRAN default).");
	else if(layout==ROW_MAJOR) printf("row-major layout (C default).");
	else
	{
		print_warning("Unrecognizing layout",__func__,"Assume row-mayor layout");
		layout=ROW_MAJOR;
		printf("row-major layout (C default).");
	}

	int columns=length/rows;
	int i, j;

	double elem;
	for(i=0;i<rows;i++)
	{
		//printf("\n [row: % 4d]\t",(i+1));
		printf("\n");
		for(j=0;j<columns;j++)
		{

			if(layout==COLUMN_MAJOR) elem=matrix[i+j*rows];
			else                     elem=matrix[j+i*columns];

			printf("% 20.15f, ",elem);
		}
	}
	printf("\n");
}

