#include "TestTools.h"
#include "PrintTools.h"

#define BYTE_ALIGN_BLK 16  //for memory alignment(better performence + consistency) [see mkl documentation]
                           //use mkl_malloc()
                           //(TODO: dinamicaly set 32-byte boundaries on processors supporting Intel Advanced Vector Extensions (AVX))

void load_matrix(const char *mName, double *matrix, int mLength)
{
	FILE *fMatrix=fopen(mName,"r");

	if(fMatrix==NULL)
	{
		char why[250]; sprintf(why,"Can't open matrix file <%s>",mName);
		print_error(why,__func__,"Abort...");
		exit(1);
	}
	int count=0;
	while(!feof(fMatrix))
	{
		fscanf(fMatrix,"%lf",&matrix[count++]);
	}
	fclose(fMatrix);
	if(mLength!=count)
	{
		print_warning("Matrix dimension problem detected",__func__,"This may cause unexpected results");
	}
}
void compute_abserror(const double *m1,const double *m2,int mlength,double *error, double *minErr, double *maxErr)
{
	double *temp_error=(double*) mkl_malloc(sizeof(double)*mlength,BYTE_ALIGN_BLK);
	memset(temp_error,0,sizeof(double)*mlength);
	int index;
	char msg[250];

	vdSub(mlength,m1,m2,temp_error);
	vdAbs(mlength,temp_error,error);

	index=cblas_idamax(mlength, error,1); maxErr[0]=error[index];
	//sprintf(msg,"MaxErr in pos: %d\t\t",index);
	printf("% 20.10f\t% 7d",maxErr[0],index);
	//index=cblas_idamin(mlength, error,1); minErr[0]=error[index];
	//sprintf(&msg[strlen(msg)],"MinErr in pos: %d",index);
	//print_info(msg);

	mkl_free(temp_error);
}

void solve_LSQ_example_MKL(const char *path,const char *exmplName, int m, int n, int r)
{
	int allocatedBytes;
	int allocatedBuffers;
	char msg[250];

	//Matrices names
	char     AName[250]; sprintf(   AName,   "%s/%s_A[%d,%d]",path,exmplName,m,n);
	char     bName[250]; sprintf(   bName,   "%s/%s_b[%d,%d]",path,exmplName,m,r);
	char    QRName[250]; sprintf(  QRName,  "%s/%s_QR[%d,%d]",path,exmplName,m,n);
	char  bHAtName[250]; sprintf(bHAtName,"%s/%s_bHat[%d,%d]",path,exmplName,m,r);
	char  coefName[250]; sprintf(coefName,"%s/%s_coef[%d,%d]",path,exmplName,n,r);

	/*long pagesize = getpagesize();
	sprintf(msg,"System pagesize is %ld",pagesize);
	print_info(msg);

	allocatedBytes = mkl_mem_stat(&allocatedBuffers);
	sprintf(msg,"Before using mkl_malloc. Intel MKL Memory Management Software uses %ld bytes in %d buffers",(long)allocatedBytes,allocatedBuffers);
	print_info(msg);
	 */
	//Allocate workspace
	double        *A=(double*) mkl_malloc(sizeof(double)*m*n,BYTE_ALIGN_BLK);
	double        *b=(double*) mkl_malloc(sizeof(double)*m*r,BYTE_ALIGN_BLK);
	double       *QR=(double*) mkl_malloc(sizeof(double)*m*n,BYTE_ALIGN_BLK);
	double     *bHat=(double*) mkl_malloc(sizeof(double)*m*r,BYTE_ALIGN_BLK);
	double     *coef=(double*) mkl_malloc(sizeof(double)*n*r,BYTE_ALIGN_BLK);
	double   *QR_MKL=(double*) mkl_malloc(sizeof(double)*m*n,BYTE_ALIGN_BLK);
	double *bHat_MKL=(double*) mkl_malloc(sizeof(double)*m*r,BYTE_ALIGN_BLK);
	double    *errQR=(double*) mkl_malloc(sizeof(double)*m*n,BYTE_ALIGN_BLK);
	double  *errbHat=(double*) mkl_malloc(sizeof(double)*m*r,BYTE_ALIGN_BLK);
	double  *errCoef=(double*) mkl_malloc(sizeof(double)*n*r,BYTE_ALIGN_BLK);

	if(A==NULL || b==NULL || QR==NULL || bHat==NULL || coef==NULL || QR_MKL==NULL || bHat_MKL==NULL || errQR==NULL || errbHat==NULL || errCoef==NULL)
	{
		print_error("Can't allocate memory for matrices",__func__,"Abort...");
		//Deallocate workspace
		mkl_free(A);
		mkl_free(b);
		mkl_free(QR);
		mkl_free(bHat);
		mkl_free(coef);
		mkl_free(QR_MKL);
		mkl_free(bHat_MKL);
		mkl_free(errQR);
		mkl_free(errbHat);
		mkl_free(errCoef);
		exit(1);
	}
	//load MATLAB Precal. values
	load_matrix(   AName,   A,m*n);
	load_matrix(   bName,   b,m*r);
	load_matrix(  QRName,  QR,m*n);
	load_matrix(bHAtName,bHat,m*r);
	load_matrix(coefName,coef,n*r);

	/*allocatedBytes = mkl_mem_stat(&allocatedBuffers);
	sprintf(msg,"After allocate & load all matrices. Intel MKL Memory Management Software uses %ld bytes in %d buffers",(long)allocatedBytes,allocatedBuffers);
	print_info(msg);*/

	memcpy(  QR_MKL,A,sizeof(double)*m*n);
	memcpy(bHat_MKL,b,sizeof(double)*m*r);

	/*allocatedBytes = mkl_mem_stat(&allocatedBuffers);
	sprintf(msg,"Before dgels (workspace query). Intel MKL Memory Management Software uses %ld bytes in %d buffers",(long)allocatedBytes,allocatedBuffers);
	print_info(msg);*/

	int info=0;
	int lwork=0;
	double wkopt=0.0;
	double* work;

	//Determining the optimal block size for workspace
	lwork = -1;
	dgels("N",&m,&n,&r,QR_MKL,&m,bHat_MKL,&m,&wkopt,&lwork,&info);

	if(info)
	{
		print_warning("Problem determining optimal block size",__func__,"Using minimum. Can cause unexpected behavior");
		lwork = 2*n*64;  //LWORK >= max(1, MN + max( MN, NRHS)*NB) , MN = min(M,N) [see dgels documentation]
	}
	else lwork = (int)wkopt;

	/*sprintf(msg,"work_length: %d",lwork);
	print_info(msg);*/

	//Allocate workspace
	work = (double*)mkl_malloc(sizeof(double)*lwork,BYTE_ALIGN_BLK);
	memset(work,0,sizeof(double)*lwork);
	//Solve LSQ problem using dgels (Intel MKL library)
	dgels("N",&m,&n,&r,QR_MKL,&m,bHat_MKL,&m,work,&lwork,&info);
	//Deallocate workspace
	mkl_free(work);

	//Check for the full rank
	if( info > 0 )
	{
		print_error("Can't compute LSQ solution. Design Matrix is singular or nearly singular",__func__,"Abort");
		exit(1);
	}

	//copy respective coefficients
	double coef_MKL[n*r];
	memcpy(    coef_MKL,    bHat_MKL,sizeof(double)*n);  //1 ROW
	//memcpy(&coef_MKL[N],&bHat_MKL[M],sizeof(double)*N);  //2 ROWS

	double alpha=1.0, beta=0.0;
	memset(bHat_MKL,0,sizeof(double)*m);
	dgemm("N","N",&m,&r,&n,&alpha,A,&m,coef_MKL,&n,&beta,bHat_MKL,&m);

	//print_info("Computing errors...");
	double maxErr=0, minErr=0;

	compute_abserror(  QR,  QR_MKL,m*n,  errQR,&minErr,&maxErr);
	//sprintf(msg,"QR Matrix	->	maxError: % 20.15f\tminError: % 20.15f",maxErr,minErr);
	//print_info(msg);
	//print_matrix(  "QR_error",  errQR,m*n,m,COLUMN_MAJOR);

	compute_abserror(bHat,bHat_MKL,m*r,errbHat,&minErr,&maxErr);
	//sprintf(msg,"bHat Matrix	->	maxError: % 20.15f\tminError: % 20.15f",maxErr,minErr);
	//print_info(msg);
	//print_matrix("bHat_error",errbHat,m*r,m,COLUMN_MAJOR);

	compute_abserror(coef,coef_MKL,n*r,errCoef,&minErr,&maxErr);
	//sprintf(msg,"coef Matrix	->	maxError: % 20.15f\tminError: % 20.15f",maxErr,minErr);
	//print_info(msg);
	//print_matrix("coef_error",errCoef,n*r,n,COLUMN_MAJOR);

	//Deallocate workspace
	mkl_free(A);
	mkl_free(b);
	mkl_free(QR);
	mkl_free(bHat);
	mkl_free(coef);
	mkl_free(QR_MKL);
	mkl_free(bHat_MKL);
	mkl_free(errQR);
	mkl_free(errbHat);
	mkl_free(errCoef);

	mkl_free_buffers();
	allocatedBytes = mkl_mem_stat(&allocatedBuffers);
	if (allocatedBytes > 0) {
		print_error("MKL memory leak",__func__,"Abort...");
		sprintf(msg,"After mkl_free_buffers. Intel MKL Memory Management Software uses %ld bytes in %d buffers",(long)allocatedBytes,allocatedBuffers);
		print_info(msg);
		exit(1);
	}
	else
	{
		/*sprintf(msg,"After free all mkl buffers. Intel MKL Memory Management Software uses %ld bytes in %d buffers",(long)allocatedBytes,allocatedBuffers);
		print_info(msg);
		print_info("No MKL memory leaks");*/
	}
}
